/*-----------------------------------------------------------------------
**  PsLeos1.prg
**
**  Example GAUSS program for estimating implied distribution from option
**  prices as in Sderlind and Svensson, "New Techniques to Extract
**  Market Expectations from Financial Instruments," CEPR D.P. 1556 and
**  NBER W.P. 5877.
**
**  The program requires the CO (Constrained Optimization) Module,
**  which is available from Aptech.
**
**  Most of the calculations are done in a number of procedures
**  in the file PsOption.prc.
**
**  This example uses LIFFE options on Bunds futures (June 1994),
**  traded on April 6, 1994.
**
**  The results are written to the file PsLeos1.out.
**
**
**
**  Paul Soderlind (nepso@hhs.se), April 10, 1997.
-----------------------------------------------------------------------*/


library co,pgraph;
#include co.ext;            /*globals for co*/
coset;

#include PsOption.prc;      /*PSLEOS procs: pricing of options, lossfn*/

output file = PsLeos1.out reset;   /*ASCII file with results*/
/*-----------------------------------------------------------------------*/

/*
**  LIFFE Bunds option data for trade date April 6, 1994
**
**  The variable names (g_X,g_C,g_F,g_k,g_r) are used as global
**  symbols and are referenced by the procedures (found in PsOption.prc)
**  to calculate the difference between actual prices and prices implied
**  by theory and a certain parameter vector.
**
**
**
*/

g_X = {                        /*strike prices, Mx1 vector*/
  92.00,  94.00,  94.50,  95.00,  95.50,  96.00,  96.50,  97.00,
  97.50,  98.00,  98.50,  99.00,  99.50,  100.0,  100.5,  101.0,
  101.5,  102.0,  102.5,  103.0,  103.5 };

g_C = {                        /*call prices, Mx1 vector*/
    5.13,    3.25,    2.83,    2.40,    2.00,    1.64,    1.31,    1.02,
   0.770,   0.570,   0.400,   0.280,   0.190,   0.130,  0.0800,  0.0500,
  0.0400,  0.0300,  0.0200,  0.0100,  0.0100 };


g_F = 97.05;                /*Forward price, scalar or Mx1 vector*/
g_k = 48/365;               /*time to expiry in years, scalar*/
g_r = 0;                    /*Interest rate: LIFFE margining=>no discounting*/

g_W = eye(rows(g_F)+rows(g_X));         /*weight matrix for price errors*/
/*-----------------------------------------------------------------------*/

/*
**  One state: solving for mean and standard deviation of distribution.
**
**  Uses: forward rate and option data.
**
**  The order of the parameters are given in the comment to _co_Bounds below.
*/



par1 = 4.575|0.03|1;         /*Initial parameter guess: mean|Std|Prob*/

coset;
_co_Active = { 1,1,0 };                 /*n=1: no probability estimated*/
_co_Bounds = { -1e256 1e256,            /* < mean  < */
                0.001 1000,             /* < std   < */
                1 1        };           /* one state => Probability=1 */
_co_GradProc = &dExOptionDist;          /*proc analytical grad of loss fn*/
_co_Options = { brent newton screen };  /*optimization method*/
_co_DirTol = 1e-5;                      /*convergence criterion*/
_co_MaxIters = 500;                     /* max number of iterations*/

{ par1,loss1,jac,retcode1 } = co( &ExOptionDist,par1 );    /*optimizing*/
"Estimated parameters for n=1: ";;par1';"";
"Loss function value: ";;loss1;
/*-----------------------------------------------------------------------*/

/*
**  Two states: solving for 2 means, 2 standard deviations, and
**  one probability of the first normal distribution.
**
**  Estimation:  (a) Assuming a probability and then optimizing wrt.
**                   mean1, mean2, std1, and std2.
**               (b) (a)is repeated for (Nprob) different values of
**                   the probability between 0.01 and 0.49. Note that
**                   there is no point in going beyond 0.5 since this
**                   is just like flipping the two distributions.
**               (c) All parameters (including the probability) are
**                   estimated simulatenously, with the intial parameter
**                   guess taken from the lowest loss function value in
**                   the loop (b).
**
**  Uses: forward rate and option data.
**
**  The order of the parameters are given in the comment to _co_Bounds below.
**
*/

coset;
_co_Active = {1,1,1,1,0};     /*loop over probability of state 1*/
_co_Bounds = { -1e256 1e256,         /* < mean1  < */
               -1e256 1e256,         /* < mean2  < */
                1e-10 1e10,          /* < std1  <  */
                1e-10 1e10,          /* < std2  <  */
                1e-10 0.49 };        /* < Prob1  < */
_co_GradProc = &dExOptionDist;       /*proc for analytical grad of loss*/
_co_Options = { one newton none };
_co_DirTol = 1e-4;
_co_MaxIters = 100;


par2 = 4.58|4.58|0.025|0.005|0.01;  /*initial parameter guess*/

NProb = 49;          /*Number of different probabilites in [0.01,0.49] to try*/

                                  /*NProb x 5 matrix to store loop results in*/
Par2M = miss( zeros(NProb,4),0 ) ~ seqa( 0.01,(0.49-0.01)/(Nprob-1),NProb );
Loss2M = miss( zeros(NProb,1),0 ); /*NProb x 1 vector to store loss fn values*/

i = 1;                              /*looping over different probabilities*/
do until i > NProb;
  par2[5] = Par2M[i,5];
  { par2,loss2,jac2,retcode2 } = co( &ExOptionDist,par2 );
  Par2M[i,.] = par2';
  Loss2M[i] = loss2;
  i;;
  i = i + 1;
endo;

MinLossIndex = minindc( Loss2M );  /*index of teration with minimum loss*/
par2 = par2M[MinLossIndex,.];      /*Parameters at minimum loss*/
loss2 = Loss2M[MinLossIndex,.];    /*minimum loss*/
"\l";"Estimated parameters for n=2 (from loop): ";;par2;
"Loss function value (from loop): ";;loss2;


_co_Active = {1,1,1,1,1};      /*estimate also Probi, initial guess from loop*/
_co_DirTol = 1e-5;
{ par2b,loss2b,jac2b,retcode2b } = co( &ExOptionDist,par2 );
"\l";"Estimated parameters for n=2 (simultaneous) : ";;par2b';
"Loss function value (simultaneous): ";;loss2b;
/*-----------------------------------------------------------------------*/

                            /*plotting some results*/

g_S = seqa( 4.47,(4.68-4.47)/99, 100 ); /*global, values to calculate pdf at*/
pdf1 = pdfMixedNormal(par1 );      /*pdf with one state*/
pdf2b = pdfMixedNormal(par2b );    /*pdf with two states*/

graphset;
title("pdfs from one and two states");
xy( g_S,pdf1~pdf2b );
/*-----------------------------------------------------------------------*/

end;


