/*-----------------------------------------------------------------------
**  PsLeos2.prg
**
**  Example GAUSS program for estimating implied distribution from option
**  prices as in Sderlind and Svensson, "New Techniques to Extract
**  Market Expectations from Financial Instruments," CEPR D.P. 1556 and
**  NBER W.P. 5877.
**
**  The program requires the CO (Constrained Optimization) Module,
**  which is available from Aptech.
**
**  Most of the calculations are done in a number of procedures
**  in the file PsOption.prc.
**
**  This example uses options of Canadian dollar futures (expiry Dec 1995),
**  traded on October 27, 1995. The data is taken from Leahy and Thomas,
**  "The Sovereignity Options: The Quebec Referendum and market Views on the
**  Canadian Dollar," International Finance Discussion Paper No. 555, Board
**  of Governors of the Federal Reserve System.
**
**  The results are written to the file PsLeos2.out.
**
**  Paul Soderlind (nepso@hhs.se), April 10, 1997.
-----------------------------------------------------------------------*/

library co,pgraph;
#include co.ext;            /*globals for co*/
coset;

#include PsOption.prc;

output file = PsLeos2.out reset;   /*ASCII file with results*/
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------*/
/*
**  Canadian dollar futures (expiry Dec 1995), option data for trade date
**  October 27, 1995.
**
**  The variable names (g_Xc,g_Xp,g_Ca,g_Pa,g_f,g_k,g_r) are used as global
**  symbols and are referenced by the procedures (found in PsOption.prc)
**  to calculate the difference between actual prices and prices implied
**  by theory and a certain parameter vector.
**
*/

g_Xc = {                    /*strike price for calls*/
  69.000,  69.500,  70.000,  70.500,  71.000,  72.000,  72.500,  73.000,
  73.500,  74.000,  74.500,  75.000,  75.500,  76.000,  76.500,  77.000,
  77.500,  78.000,  78.500 };

g_Xp = {                   /*strike price for puts*/
  67.500,  68.000,  68.500,  69.000,  69.500,  70.000,  70.500,  71.000,
  71.500,  72.000,  72.500,  73.000,  73.500,  74.000,  74.500,  75.000,
  75.500,  76.000 };

g_Ca = {                  /*prices of calls*/
  4.4700,  4.0400,  3.6100,  3.2000,  2.7900,  2.0400,  1.6900,  1.4000,
  1.1500, 0.95000, 0.80000, 0.67000, 0.54000, 0.42000, 0.32000, 0.27000,
 0.21000, 0.13000,0.070000 };

g_Pa = {                   /*prices of puts*/
 0.10000, 0.15000, 0.19000, 0.24000, 0.30000, 0.37000, 0.46000, 0.55000,
 0.66000, 0.79000, 0.94000,  1.1400,  1.3900,  1.6800,  2.0300,  2.4000,
  2.7700,  3.1400 };

g_f = 73.260;             /*futures price: US cents per CAD*/
g_k = 0.11781;             /*time to expiry in years*/
g_r = 0.052100;            /*US T-bill rate*/

g_W = eye( rows(g_F)+rows(g_Xc)+rows(g_Xp) );         /*weight matrix*/
/*-----------------------------------------------------------------------*/

/*
**  One state: solving for mean and standard deviation of distribution.
**
**  Uses: forward rate and option data.
**
**  The order of the parameters are given in the comment to _co_Bounds below.
**  Probability=1 since only one state.
**  For simiplicity, w1 and w2 are not estimated for n=1, bit set to 0.5.
**
*/


par1 = 4.29|0.04|1|0.5|0.5;   /*mean|std|Probi,w1,w2*/
coset;
_co_Active = {1,1,0,0,0};                 /*no Probi,w1, or w2*/
_co_Bounds = { -1e256 1e256,              /* < mean < */
                0.001 1000,               /* < std < */
                1 1,                      /* prob state 1 = 1 */
                1e-4 0.999,               /* < w1 < */
                1e-4 0.999 };             /* < w2 < */

_co_Options = { brent newton none central };
_co_DirTol = 1e-5;
_co_MaxIters = 500;

{ par1,loss1,jac,retcode1 } = co( &AmerOptionDist,par1 );
"Estimated parameters for n=1: ";;par1';;
"Loss function value : ";;loss1;"";
/*-----------------------------------------------------------------------*/

/*
**  Two states: solving for 2 means, 2 standard deviations,
**  one probability of the first normal distribution, and w1 and w2.
**
**  Estimation: (a) w1 and w2 are set to 0.5.
**              (b) assuming a probability of state 1, loss fn is
**                  minimized wrt mean1, mean2, std1, and std2
**              (c) (b) is repeated for (Nprob) different values of
**                  the probability between 0.01 and 0.49. Note that
**                  there is no point in going beyond 0.5 since this
**                  is just like flipping the two distributions.
**
**  Uses: forward rate and option data.
**
**  The order of the parameters are given in the comment to _co_Bounds below.
**
*/

coset;
_co_Active = {1,1,1,1,0,0,0}; /*mean1,mean2,std1,std2,Prob1,w1,w2*/
_co_Bounds = { -1e256 1e256,             /* < mean1 < */
               -1e256 1e256,             /* < mean2 < */
                1e-4  1e10,              /* < std1 <; lower limit speeds up*/
                1e-4  1e10,              /* < std2 <; estimation. Relaxable*/
                1e-10 0.49,              /* < Prob state 1 < */
                1e-4  0.999,             /* < w1 < */
                1e-4  0.999 };           /* < w2 < */

_co_Options = { one newton none central };  /*central numerical gradient*/
_co_DirTol = 5e-4;                          /*is important*/
_co_MaxIters = 100;


par2 = 4.29|4.29|0.044|0.010|0.01|0.5|0.5;    /*initial guess*/

NProb = 49;         /*loop over Probi values: 0.01 to 0.49*/
Par2M = miss( zeros(NProb,4),0 ) ~ seqa( 0.01,(0.49-0.01)/(Nprob-1),NProb ) ~
        miss( zeros(NProb,2),0 );
Loss2M = miss( zeros(NProb,1),0 ); /*NProb x 1 vector to store loss fn values*/

i = 1;
do until i > NProb;
  par2[5] = Par2M[i,5];
  { par2,loss2,jac2,retcode2 } = co( &AmerOptionDist,par2 );
  Par2M[i,.] = par2';
  Loss2M[i] = loss2;
  i;;
  i = i + 1;
endo;

MinLossIndex = minindc( Loss2M );  /*index of teration with minimum loss*/
par2 = par2M[MinLossIndex,.];      /*Parameters at minimum loss*/
loss2 = Loss2M[MinLossIndex,.];    /*minimum loss*/
"\l";"Estimated parameters for n=2 (from loop): ";;par2;
"Loss function value (from loop): ";;loss2;"";
/*-----------------------------------------------------------------------*/

/*
**  Three states: solving for 3 means, 3 standard deviations,
**  two probabilites of the first two normal distributions, and w1 and w2.
**
**  Estimation: (a) w1 and w2 are set to 0.5.
**              (b) assuming probabilities of states 1 and 2, loss fn is
**                  minimized wrt mean 1-3, and std 1-3
**              (c) (b) is repeated for (Nprob) different sets of
**                  probabilities.
**
**  Uses: forward rate and option data.
**
**  The order of the parameters are given in the comment to _co_Bounds below.
**
**  The probability grid requires some thought in order to be efficient.
**  Note that all 6 permutations of the probabilities (0.2,0.3,0.5) must
**  gives the same loss function value. For instance, (0.2,0.3,0.5)
**  and (0.2,0.5,0.3) combined with flipping (mean2,std2,mean3,std3) to
**  (mean3,std3,mean2,std2) gives the same distribution but with the order
**  of the two last normals interchanged.  We can therefore limit the grid
**  to the case where a1 >= a2 >= a3 >= 0  (in practice a1 > a2 > a3 >0).
**
*/

coset;
_co_Active = {1,1,1,1,1,1,0,0,0,0}; /*mean1-3,std1-3,Prob1-2,w1,w2*/
_co_Bounds = { -1e256 1e256,           /* < mean1 < */
               -1e256 1e256,           /* < mean2 < */
               -1e256 1e256,           /* < mean3 < */
                1e-4  1e10,            /* < std1 <; lower limit speeds up*/
                1e-4  1e10,            /* < std2 <  estimation. Could be  */
                1e-4  1e10,            /* < std3 <  relaxed */
                1e-10 0.999,           /* < prob state 1 < */
                1e-10 0.999,           /* < prob state 2 < */
                1e-4  0.999,           /* < w1 < */
                1e-4  0.999 };         /* < w2 < */

_co_Options = { one newton none central };  /*central numerical gradient*/
_co_DirTol = 1e-4;                          /*is important*/
_co_MaxIters = 100;


par3 = 4.27|4.29|4.35|0.05|0.01|0.024|0.44|0.34|(1-0.83)|(1-0.46);
par3 = 4.29|4.29|4.05|0.04|0.01|0.2|0.95|0.04|0.5|0.5;

Prob12 = Prob123Setup( seqa(0.02,0.01,96),0.4|0|0,0.9|1|1 ); /*smart grid*/
Nprob = rows(Prob12);

Par3M = miss( zeros(NProb,6),0 ) ~ Prob12 ~  miss( zeros(NProb,2),0 );
Loss3M = miss( zeros(NProb,1),0 ); /*NProb x 1 vector to store loss fn values*/

i = 1;                                            /*fval,tcode*/
do until i > NProb;
  par3[7:8] = Prob12[i,.]';
  { par3,loss3,jac3,retcode3 } = co( &AmerOptionDist,par3 );
  Par3M[i,.] = par3';
  Loss3M[i] = loss3;
  i;;
  i = i + 1;
endo;

MinLossIndex = minindc( Loss3M );  /*index of teration with minimum loss*/
par3 = par3M[MinLossIndex,.];      /*Parameters at minimum loss*/
loss3 = Loss3M[MinLossIndex,.];    /*minimum loss*/
"\l";"Estimated parameters for n=3 (from loop): ";;par3;
"Loss function value (from loop): ";;loss3;"";


_co_Active = {1,1,1,1,1,1,1,1,0,0};       /*estimate also probs*/
_co_C = { 0 0 0 0 0 0 -1 -1 0 0 };         /*a1+a2<1 <=> -a1 - a2 > -1*/
_co_D = -1;                                /*_co_C*pars >= _co_D*/
_co_DirTol = 5e-5;
_co_MaxIters = 250;
_co_Options = { one newton screen central };  /*central numerical gradient*/
{ par3b,loss3b,jac3b,retcode3b } = co( &AmerOptionDist,par3 );
"\l";"Estimated parameters for n=3 (all except w1 and w2): ";;par3b;
"Loss function value: ";;loss3b;"";

_co_Active = {1,1,1,1,1,1,1,1,1,1};       /*estimate all*/
{ par3c,loss3c,jac3c,retcode3c } = co( &AmerOptionDist,par3b );
"\l";"Estimated parameters for n=3 (all): ";;par3c;
"Loss function value: ";;loss3c;"";
/*-----------------------------------------------------------------------*/

                                        /*plotting results*/
g_S = seqa( 4.15,(4.4-4.15)/99, 100 ); /*global, values to calculate pdf at*/
pdf1  = pdfMixedNormal( par1[1:3] );       /*pdf with one state*/
pdf2  = pdfMixedNormal( par2[1:5]' );      /*pdf with two states*/
pdf3c = pdfMixedNormal( par3c[1:8] );      /*pdf with three states*/

graphset;
title("pdfs from one, two, and three states");
xy( g_S,pdf1~pdf2~pdf3c );
/*-----------------------------------------------------------------------*/

end;





/*Procedures*/
/*-----------------------------------------------------------------------
**  Prob123Setup
**
**  for generating grid of probability values for n=3 in
**  a smart way. The probabilities for state 1, 2, and 3 (a1,a2,a3)
**  fullfill:    (a) a1,a2,a3 > 0
**               (b) a1 > a2 > a3 (since permutations give same result)
**               (c) a3 = 1 - a1 - a2
**               (d) ProbMin[i] <= ai <= ProbMax[i], where the 3x1
**                   vectors are supplied by the user.
**
**
**
**
**  Used in Quebec2.prg.
**
**
**  Paul Soderlind 1996-08-19  9:36:14
-----------------------------------------------------------------------*/

proc Prob123Setup( Prob1,ProbMin,ProbMax );

  local N,Prob12,Prob123,logcon;

  Prob1 = rev( sortc(Prob1,1) );        /*type 0.98,0.97,...,0.02*/
  N = rows(Prob1);

  Prob12 = (Prob1.*.ones(N,1)) ~ (ones(N,1).*.Prob1);
  Prob123 = Prob12 ~ (1-sumc(Prob12')); /*probabilities for states 1,2, and 3*/

  logcon = dotfle( Prob123[.,3],0 ) .or             /*a3 <= 0 */
           dotfge( Prob123[.,3],Prob123[.,2] ) .or  /*a3 >= a2 */
           dotfge( Prob123[.,2], Prob123[.,1] );    /*a2 >= a1 */
  logcon  = indexcat( logcon,0 );        /* index for a1 > a2 > a3 > 0*/
  Prob123 = Prob123[logcon,.];          /*pick out those cases*/

  logcon =  dotflt( Prob123,ProbMin' ) .or      /*Prob < ProbMin*/
            dotfgt( Prob123,ProbMax' );         /*Prob > ProbMax*/
  logcon =  sumc(logcon') .>= 1;
  logcon  = indexcat( logcon,0 );
  Prob123 = Prob123[logcon,.];          /*pick out those cases*/

  retp( Prob123[.,1:2] );

  endp;
/*-----------------------------------------------------------------------*/



