/*-----------------------------------------------------------------------
**  PsOption.prc
**
**  procs for pricing European options and fitting normal
**  distributions to them.
**
**
**
**
**
**  Paul Soderlind 1996-08-14, revised March 1997.
-----------------------------------------------------------------------*/


declare g_W ?= .;          /*Declaring globals used in the procs*/
declare g_C ?= .;
declare g_F ?= .;
declare g_X ?= .;
declare g_k ?= .;
declare g_r ?= .;
declare g_S ?= .;

declare g_Pa ?= .;
declare g_Ca ?= .;
declare g_w1 ?= .;
declare g_w2 ?= .;
declare g_Xc ?= .;
declare g_Xp ?= .;


/******************************* Black-Scholes *************************/

/*-----------------------------------------------------------------------
**  ExOption1( StdS )
**
**  Purpose:   Calculates call price of option written on futures,
**             for a given standard deviation of underlying asset.
**
**  Input:     Stds   --  scalar, standard deviation of underlying asset,
**                        corresponding to sigma*sqrt(g_k) in B&S)
**
**  Global:    g_F    --   scalar, forward price
**  (inputs)   g_X    --   Mx1 vector, strike prices
**             g_k    --   scalar, time to expiry
**             g_r    --   scalar, interest rate per period
**
**
**
**  Output:    C      --  Mx1 vector, implied call prices.
**
**
**  Note:      Uses Black's formula for an option on a futures contract.
**
**
**  Paul Soderlind 9607
-----------------------------------------------------------------------*/

proc ExOption1( StdS );

  local d1,d2,C;

  StdS = abs(StdS);


  d1 = ( ln(g_F./g_X) + StdS^2/2 )/StdS;
  d2 = ( ln(g_F./g_X) - StdS^2/2 )/StdS;
  C = g_F.*exp(-g_k*g_r) .*cdfn(d1) -
      g_X      .*exp(-g_k*g_r) .*cdfn(d2);

  retp( C );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  dExOption1( StdS )
**
**  Purpose:   Analytical gradient of call prices in ExOption1.
**
**  Usage:      dC_dStds = dExOption1( StdS );
**
**  See ExOption1 for details.
**
**
**  Paul Soderlind 9607
-----------------------------------------------------------------------*/

proc dExOption1( StdS );

  local d1,d2,dd1,dd2,C;

  StdS = abs(StdS);

  d1 = ( ln(g_F./g_X) + StdS^2/2 )/StdS;
  d2 = ( ln(g_F./g_X) - StdS^2/2 )/StdS;

  dd1 =  1  - ( ln(g_F./g_X) + StdS^2/2 )/StdS^2;
  dd2 = -1  - ( ln(g_F./g_X) - StdS^2/2 )/StdS^2;


  C = g_F.*exp(-g_k*g_r) .*pdfn(d1).*dd1 -
      g_X      .*exp(-g_k*g_r) .*pdfn(d2).*dd2;

  retp( C );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  ExOption1b
**
**  Purpose:   loss function for option prices in ExOption1 (Black's formula):
**             squared pricing errors time 10000.  Used for extracting implied
**             volatility.
**
**  Usage:     loss_value = ExOption1b( StdS );
**
**  See ExOption1 for details.
**
**  Paul Soderlind 9607
-----------------------------------------------------------------------*/

proc ExOption1b( StdS );

  local C;

  C = ExOption1( StdS );

  retp( 10000*(C-g_C)^2 );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  dExOption1b
**
**  Purpose:   Analytical gradient of the loss function in dExOption1b.
**
**  Usage:     dLoss_dStds = dExOption1b( StdS );
**
**  See ExOption1 for details.
**
**  Paul Soderlind 9607
-----------------------------------------------------------------------*/

proc dExOption1b( StdS );

  local C,dLoss;

  C = ExOption1( StdS );
  dLoss = 10000*2*(C-g_C) * dExOption1( StdS );

  retp( dLoss );

endp;
/*-----------------------------------------------------------------------*/

/****************************** End Black-Scholes ************************/





/************************* Mixture of log-normals  ***********************/

/*-----------------------------------------------------------------------
**  Call1
**
**  Purpose:   Proc for generic option price formula, using -kr=Em+Var(m)/2,
**             but not asset pricing equation for underlying asset,
**             n different states (mixture of n lognormal distributions).
**             Calculates call prices for M different strike prices.
**
**  Usage:     C = proc Call1( k,r,sbar_ms,SigmaSS,X,Probi );
**
**  Inputs:    k        --   scalar, periods to expiry date  (could be <1 );
**             r        --   scalar, interest rate per period
**             sbar_ms  --   nx1 vector of E(t)lnS(t+k) +
**                           Cov(t)[lnM(t+k),lnS(t+k)]
**             SigmaSS  --   nx1 vector of Var(t)[lnS(t+k)]
**             X        --   Mx1 vector of strike prices
**             Probi    --   nx1 vector of probabilities of different states
**
**  Output:    C        --   Mx1 vector of call prices at M strike prices
**
**
**
**  Paul Soderlind 1996-06-06.
-----------------------------------------------------------------------*/

proc Call1( k,r,sbar_ms,SigmaSS,X,Probi );

  local d1,d2,C;

  d1 = ( sbar_ms + SigmaSS - ln(X') )./sqrt(SigmaSS);
  d2 = ( sbar_ms -           ln(X') )./sqrt(SigmaSS);

  C = exp( -k*r )*exp( sbar_ms + SigmaSS/2 ) .* cdfn(d1)  -
      exp( -k*r )*X'.*cdfn(d2);

  C = sumc( Probi.*C );     /*expectation over states*/

  retp( C );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  AssetPriceMix1
**
**  Purpose:   General proc for asset pricing equation with n states .
**             Calculates one asset price.
**
**  Usage:     F = AssetpriceMix1( k,r,sbar_ms,SigmaSS,Probi );
**
**  Inputs:    see Call1
**
**  Output:    F    -- scalar, asset price
**
**
**  Application:    forward exchange rate:  k=0,r=0.
**
**
**  Paul Soderlind 1996-06-07
-----------------------------------------------------------------------*/

proc AssetpriceMix1( k,r,sbar_ms,SigmaSS,Probi );

  local S;

  S = exp(-k*r) * sumc(  Probi .* exp(sbar_ms + SigmaSS/2)  );

  retp( S );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  ExOptionDist
**
**  Purpose:   Loss function for fitting n log-normal distributions
**             from option data. Uses AssetpriceMix1 and Call1. The
**             proc calculates the weighted squared pricing errors of
**             the forward price and the option prices.
**
**  Usage:     loss_value = ExOptionDist( pars );
**
**  Input:     pars    --  n*3-1 vector of parameters in log-normal
**                         distributions in the following order:
**                         mean_1,mean_2,...,mean_n,
**                         Std_1 ,Std_2 ,...,Std_n,
**                         Prob_1 ,Prob_2 ,...,Prob_n-1
**
**  Globals:
**  Global:    g_F    --   scalar or Mx1 vector, forward price. If Mx1,
**  (inputs)               then it is one forward price for each call price.
**             g_X    --   Mx1 vector, strike prices
**             g_C    --     ""      , call option prices
**             g_k    --   scalar, time to expiry
**             g_r    --   scalar, interest rate per period
**             g_W    --   (1+M)x(1+M) or (M+M)x(M+M) weighting matrix
**
**
**  Output:    loss_value  the value of
**                 [(F-g_F)|(C-g_C)]'g_W*[(F-g_F)|(C-g_C)]
**
**
**  Note:      Prob_n is given by 1-sum(Prob_1,..,Prob_n-1)
**
**
**
**  Paul Soderlind 1996-06-07
-----------------------------------------------------------------------*/

proc ExOptionDist( pars );

  local n,sbar_ms,SigmaSS,Probi,C,F,dev,loss;

  n = floor(  (rows(pars) + 1)/3  );               /*#states*/
  sbar_ms = pars[1:n];
  SigmaSS = pars[n+1:2*n]^2;     /*estimating standard deviation*/
  if n > 1;
    Probi = pars[2*n+1:3*n-1];
    Probi = Probi|(1-sumc(Probi));      /*Probs sum to one*/
  else;
    Probi = pars[2*n+1:2*n+1];          /*n=1: one distribution*/
  endif;


  C = Call1( g_k,g_r,sbar_ms,SigmaSS,g_X,Probi );
  F = AssetpriceMix1( 0,0,sbar_ms,SigmaSS,Probi );  /*Forward rate*/

  dev = (F-g_F)|(C-g_C);
  loss = dev'g_W*dev;

  retp( 10000*loss );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  dExOption
**
**  Purpose:   Analytical gradient of future|call prices and of
**             loss function in ExOptionDist
**
**  Usage:     { ddev_dpar,dloss_dpar } = dExOption( pars );
**
**  Input:     See ExOptionDist
**
**  Globals:   See ExOptionDist
**
**  Outputs:   ddev_dpar    --   (1+M)xNpar or (M+M)xNpar matrix, gradients
**                               of (1 or M) futures and (M) call prices
**             dloss_dpar   --   1xNpar vector, gradient of loss function
**
**
**  Paul Soderlind 1996-08-14 15:17:48
-----------------------------------------------------------------------*/

proc (2) = dExOption( pars );

  local n,sbar_ms,SigmaSS,Probi,C,F,dev,
        dF_dPar,dF_dsbar,dF_dSigma,dF_dProbi,
        d1,d2,C1,C2,dC_dPar,dC_dsbar,dC_dSigma,dC_dProbi,
        ddev_dpar,dloss_dpar;


  n = floor(  (rows(pars) + 1)/3  );               /*#states*/
  sbar_ms = pars[1:n];
  SigmaSS = pars[n+1:2*n]^2;     /*estimating standard deviation*/
  if n > 1;
    Probi = pars[2*n+1:3*n-1];
    Probi = Probi|(1-sumc(Probi));      /*Probs sum to one*/
  else;
    Probi = pars[2*n+1:2*n+1];          /*n=1: one distribution*/
  endif;


  C = Call1( g_k,g_r,sbar_ms,SigmaSS,g_X,Probi ); /*prices*/
  F = AssetpriceMix1( 0,0,sbar_ms,SigmaSS,Probi );  /*Forward exc rate*/
  dev = (F-g_F)|(C-g_C);


  dF_dsbar  = Probi .* exp(sbar_ms + SigmaSS/2);    /*dFuture price/dparams*/
  dF_dSigma = Probi .* exp(sbar_ms + SigmaSS/2).*sqrt(SigmaSS);
  dF_dProbi =          exp(sbar_ms + SigmaSS/2);
  if n > 1;
    dF_dProbi[1:n-1] = dF_dProbi[1:n-1]- dF_dProbi[n];
    dF_dProbi = trimr(dF_dProbi,0,1);   /*last Probi defined by sum*/
    dF_dPar = exp(-g_k*g_r) * (dF_dsbar'~dF_dSigma'~dF_dProbi');
  else;
    dF_dPar = exp(-g_k*g_r) * (dF_dsbar'~dF_dSigma'~dF_dProbi');
  endif;
  dF_dPar = ones(rows(g_F),1) * dF_dpar; /*dF_dpar is scalar, here transformed*/
                                         /*to same length as g_F (1x1 or Mx1)*/


  d1 = ( sbar_ms + SigmaSS - ln(g_X') )./sqrt(SigmaSS);  /*dcall/dparams*/
  d2 = ( sbar_ms -           ln(g_X') )./sqrt(SigmaSS);
  C1 = exp( sbar_ms + SigmaSS/2 ) .* cdfn(d1);
  C2 = g_X'.*cdfn(d2);                /*C=exp(-k*r)[C1-C2]*/
  dC_dsbar = C1 + exp( sbar_ms+SigmaSS/2 ) .* pdfn(d1) ./ sqrt(SigmaSS) -
             g_X' .* pdfn(d2) ./ sqrt(SigmaSS);
  dC_dsbar = Probi.*dC_dsbar;
  dC_dSigma = C1.*sqrt(SigmaSS) +
             exp( sbar_ms+SigmaSS/2 ) .* pdfn(d1) .*
               ( 2 - d1./sqrt(SigmaSS )) -
             g_X' .* pdfn(d2) .* ( -d2./sqrt(SigmaSS) );
  dC_dSigma = Probi.*dC_dSigma;
  dC_dProbi = C1 - C2;
  if n > 1;
    dC_dProbi[1:n-1,.] = dC_dProbi[1:n-1,.]- dC_dProbi[n,.];
    dC_dProbi = trimr(dC_dProbi,0,1);   /*last Probi defined by sum*/
    dC_dPar = exp(-g_k*g_r) * (dC_dsbar|dC_dSigma|dC_dProbi)';
  else;
    dC_dPar = exp(-g_k*g_r) * (dC_dsbar|dC_dSigma|dC_dProbi)';
  endif;

  ddev_dpar = dF_dpar|dC_dpar;          /*gradient of future&call prices*/
  dloss_dpar = 2*(ddev_dpar)'g_W*dev;   /*gradient of loss function*/

  retp( ddev_dpar,10000*dloss_dpar );


endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  dExOptionDist
**
**  Purpose:   Analytical gradient of loss function in ExOptionDist.
**             Picks out a subset of the results from dExOption.
**
**  Usage:     dloss_dpar = dExOptionDist( pars );
**
**  See dExOption for details.
**
**
**  Paul Soderlind 1996-08-14 15:18:30
-----------------------------------------------------------------------*/

proc dExOptionDist( pars );


  local ddev_dpar,dloss_dpar;

  { ddev_dpar,dloss_dpar } =  dExOption( pars );

  retp( dloss_dpar );


endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  dExOptionFC
**
**  Purpose:   Analytical gradaient of futures and call prices.
**             Picks out a subset of the results from dExOption.
**
**  Usage:     ddev_dpar = dExOptionFC( pars );
**
**
**  See dExOption for details.
**
**  Paul Soderlind 1996-08-14 15:18:30
-----------------------------------------------------------------------*/

proc dExOptionFC( pars );


  local ddev_dpar,dloss_dpar;

  { ddev_dpar,dloss_dpar } =  dExOption( pars );

  retp( ddev_dpar );


endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  AmerOptionDist
**
**  Purpose:   Loss function for fitting n log-normal distributions from
**             American-style option data: uses AssetpriceMix1 and Call1,
**             using the technique in Melick and Thomas.
**
**  Usage:     loss_value = AmerOptionDist( pars );
**
**  Input:     pars    -   as in ExOptionDist, plus (comes last)
**                       w1  (weight on lower bound for ITM options)
**                       w2  (weight on lower bound for OTM options)
**
**
**  Global:    g_F    --   scalar forward price.
**  (inputs)   g_Xc    --   M1x1 vector, strike prices for calls
**             g_Xp    --   M2x1 vector, strike prices for puts
**             g_Ca    --   M1x1 call option prices
**             g_Pa    --   M2x1 put option prices
**             g_k    --   scalar, time to expiry
**             g_r    --   scalar, interest rate per period
**             g_W    --   (1+M1+M2)x(1+M1+M2) weighting matrix
**
**  Output:    loss_value  the value of
**             [(F-g_F)|(C-g_Ca)|(P-g_Pa)]'g_W*[(F-g_F)|(C-g_Ca)|(P-g_Pa)]
**
**
**  Note:      Prob_n is given by 1-sum(Prob_1,..,Prob_n-1)
**
**  See ExOptionDist for details.
**
**  Paul Soderlind 1996-06-07
-----------------------------------------------------------------------*/

proc AmerOptionDist( pars );

  local n,sbar_ms,SigmaSS,Probi,C,F,CXp,P,CallLow,CallHigh,PutLow,
        PutHigh,wcall,wput,dev,loss,vv,w1,w2;

  n = floor(  (rows(pars) + 1-2)/3  );               /*#states*/
  sbar_ms = pars[1:n];
  SigmaSS = pars[n+1:2*n]^2;     /*estimating standard deviation*/
  if n > 1;
    Probi = pars[2*n+1:3*n-1];
    Probi = Probi|(1-sumc(Probi));      /*Probs sum to one*/
    vv = 3*n-1;
  else;
    Probi = pars[2*n+1:2*n+1];          /*n=1: one distribution*/
    vv = 2*n+1;
  endif;
  w1 = pars[vv+1];                     /*weight on in-the-money options*/
  w2 = pars[vv+2];


  C = Call1( g_k,g_r,sbar_ms,SigmaSS,g_Xc,Probi );  /*C at X for calls*/
  F = AssetpriceMix1( 0,0,sbar_ms,SigmaSS,Probi );  /*Forward rate*/

  CXp = Call1( g_k,g_r,sbar_ms,SigmaSS,g_Xp,Probi ); /*C at X for puts*/
  P = CXp + exp(-g_k*g_r)*(g_Xp-g_F);    /*put-call parity for futures*/

  CallLow = maxc(  ( (F-g_Xc)~C )'  );         /*Bounds a la Mellick&Thomas*/
  CallHigh = exp(g_k*g_r)*C;
  PutLow = maxc(  ( (g_Xp-F)~P )'  );
  PutHigh = exp(g_k*g_r)*P;

  wcall = g_Xc .<= g_F;                  /*weights of lower&upper bound*/
  wcall = wcall*w1 + (1-wcall)*w2;  /*w1 if X<=F for call (in-the-money)*/
  wput = g_Xp .>= g_F;
  wput = wput*w1 + (1-wput)*w2;     /*w1 if X>=F for put (in-the-money)*/

  C = wcall.*CallLow + (1-wcall).*CallHigh;
  P = wput.*PutLow   + (1-wput).*PutHigh;


  dev = (F-g_F)|(C-g_Ca)|(P-g_Pa);
  loss = dev'g_W*dev;

  retp( 10000*loss );

endp;
/*-----------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
**  pdfMixedNormal
**
**  Purpose:   Calculating pdf of mixture of normal distributions
**
**  Usage:     pdf = pdfMixedNormal( pars );
**
**  Input:     pars    --  see ExOptionDist for details.
**
**  Global:    g_S     --  Mx1 vector, values to calculate pdf at
**  (input)
**
**  Output:    pdf     --  Mx1 vector, pdf(g_S;pars)
**
**
**
**  Paul Soderlind 960612
-----------------------------------------------------------------------*/

proc pdfMixedNormal( pars );

  local n,meanM,StdM,Probi,pdf;


  n = floor(  (rows(pars) + 1)/3  );               /*#states*/
  meanM = pars[1:n];
  StdM = pars[n+1:2*n];
  if n > 1;
    Probi = pars[2*n+1:3*n-1];
    Probi = Probi|(1-sumc(Probi));      /*Probs sum to one*/
  else;
    Probi = pars[2*n+1:2*n+1];          /*n=1: one distribution*/
  endif;


  pdf = pdfn( (g_S-meanM')./StdM' )./StdM';
  pdf = sumc( pdf' .* Probi );

  retp(pdf);

endp;
/*-----------------------------------------------------------------------*/

/************************ End mixture of log-normals *********************/




